'use strict';

angular.module('app')
    .service('BungieClient', function($q, $http, ApiResponse, DSCacheFactory) {
        var RECENT_SIZE = 10;

        var BUNGIE_BASE_URL = 'http://www.bungie.net/Platform/Destiny/';
        var ENDPOINT_CHARACTER_LIST = '{platform}/Account/{memberId}/?definitions=true';
        var ENDPOINT_CHARACTER_INVENTORY = '{platform}/Account/{memberId}/Character/{characterId}/Inventory/?definitions=true';
        var ENDPOINT_CHARACTER_PROGRESSION = '{platform}/Account/{memberId}/Character/{characterId}/Progression/?definitions=true';
        var ENDPOINT_CHARACTER_ACTIVITIES = '{platform}/Account/{memberId}/Character/{characterId}/Activities/?definitions=true';
        var ENDPOINT_CHARACTER_ACTIVITY_HISTORY = 'Stats/ActivityHistory/{platform}/{memberId}/{characterId}/?mode=0&count=250&definitions=true';
        var ENDPOINT_SEARCH_FOR_MEMBER = 'SearchDestinyPlayer/{platform}/{name}/';

        var memberNameCache = DSCacheFactory('app.bungie.client:member-name', { storageMode: 'localStorage' });
        var recentCache = DSCacheFactory('app.bungie.client:recent-list', { storageMode: 'localStorage' });
        var searchCache = DSCacheFactory('app.bungie.client:search', { storageMode: 'localStorage' });

        return {
            getRecentSearches: function() {
                return recentCache.get('data');
            },

            addRecentSearch: function(platform, memberId) {
                var cache = recentCache.get('data');

                if (cache === undefined) {
                    cache = [];
                }

                // remove existing entry of same id
                for (var i = 0; i < cache.length; i++) {
                    if (cache[i].platform == platform && cache[i].memberId == memberId) {
                        cache.splice(i, 1);
                        break;
                    }
                }

                cache.unshift({
                    platform: platform,
                    memberId: memberId,
                    memberDetails: this.getMemberDetails(platform, memberId)
                });

                if (cache.length == (RECENT_SIZE + 1)) {
                    cache.pop();
                }

                recentCache.put('data', cache);
            },

            createUri: function(endpoint, replace) {
                replace = (typeof replace == "undefined") ? [] : replace;

                angular.forEach(replace, function(value, key) {
                    endpoint = endpoint.replace('{' + key + '}', value);
                });

                return BUNGIE_BASE_URL + endpoint;
            },

            getMemberDetails: function(platform, memberId) {
                var cacheKey = [platform, memberId].join(':');
                var cache = memberNameCache.get(cacheKey);

                if (cache !== undefined) {
                    return cache;
                }

                return null;
            },

            /**
             * First, lookup the name by platform/memberId then inject the
             * displayName into the character data before resolving the promise.
             *
             * @param platform
             * @param memberId
             * @returns {*}
             */
            findCharacters: function(platform, memberId) {
                var d = $q.defer();

                this.request(ENDPOINT_CHARACTER_LIST, { memberId: memberId, platform: platform })
                    .then(function(result) {
                        if (!result.getData()) {
                            // this only happens if the bungie search api has cached someone who no longer exists
                            d.reject({ bungieSearchInvalid: true });
                        } else {
                            result.getData()['memberDetails'] = this.getMemberDetails(platform, memberId);
                            d.resolve(result);
                        }
                    }.bind(this), function(error) {
                        d.reject(error);
                    });

                return d.promise;
            },

            findCharacterInventory: function(platform, memberId, characterId) {
                return this.request(ENDPOINT_CHARACTER_INVENTORY, {
                    memberId: memberId,
                    characterId: characterId,
                    platform: platform
                })
            },

            findCharacterProgression: function(platform, memberId, characterId) {
                return this.request(ENDPOINT_CHARACTER_PROGRESSION, {
                    memberId: memberId,
                    characterId: characterId,
                    platform: platform
                })
            },

            findCharacterActivities: function(platform, memberId, characterId) {
                return this.request(ENDPOINT_CHARACTER_ACTIVITY_HISTORY, {
                    platform: platform,
                    memberId: memberId,
                    characterId: characterId
                })
            },

            findMembersByName: function(name) {
                var d = $q.defer();
                var results = [];

                this.findMemberByPlatformAndName(1, name)
                    .then(function(result) {
                        if (result) {
                            results.push(result);
                        }

                        this.findMemberByPlatformAndName(2, name)
                            .then(function(result) {
                                if (result) {
                                    results.push(result);
                                }
                                return d.resolve(results);
                            }, function(error) {
                                d.reject(error);
                            })
                    }.bind(this), function(error) {
                        d.reject(error);
                    });

                return d.promise;
            },

            findMemberByPlatformAndName: function(platform, name) {
                var d = $q.defer();

                var uri = this.createUri(ENDPOINT_SEARCH_FOR_MEMBER, {
                    name: name.toLowerCase(),
                    platform: platform
                });

                var cache = searchCache.get(uri);
                if (cache !== undefined) {
                    // missing platform results are cached as "false"
                    if (cache !== false) {
                        this.addRecentSearch(platform, cache.memberId);
                    }
                    d.resolve(cache);
                    return d.promise;
                }

                $http.get(uri)
                    .then(function(result) {
                        if (result.data.Response.length == 0) {
                            // cache search results that fail for a one hour in case the account gets created
                            searchCache.put(uri, false, { maxAge: 3600000 });
                            d.resolve();
                            return d.promise;
                        }

                        var memberId = result.data.Response[0].membershipId;

                        var details = {
                            platform: platform,
                            memberId: memberId,
                            name: result.data.Response[0].displayName
                        };

                        // save the member details in a cache for lookup retrieval
                        var detailsCacheKey = [platform, memberId].join(':');
                        memberNameCache.put(detailsCacheKey, result.data.Response[0]);
                        searchCache.put(uri, details);

                        this.addRecentSearch(platform, memberId);

                        d.resolve(details);
                        return d.promise;
                    }.bind(this), function(error) {
                        d.reject(error);
                    });

                return d.promise;
            },

            request: function(endpoint, params, cache) {
                var d = $q.defer();
                var uri = this.createUri(endpoint, params);

                cache = cache === undefined ? false : cache;

                $http.get(uri, { cache: cache })
                    .then(function(result) {
                        d.resolve(new ApiResponse(result.data));
                    }, function(error) {
                        d.reject(error);
                    });

                return d.promise;
            }
        };
    });
