'use strict';

angular.module('app')
    .service('CharacterService', function ($q, BungieClient, DSCacheFactory) {
        var factionTagMap = {
            'character_prestige': 'Prestige Levels',
            'faction_cryptarch': 'Cryptarch',
            'faction_eris': 'Eris',
            'faction_event_iron_banner': 'Iron Banner',
            'faction_event_queen': 'The Queen\'s Wrath',
            'faction_fotc_vanguard': 'Vanguard',
            'faction_pvp': 'Crucible',
            'faction_pvp_dead_orbit': 'Dead Orbit',
            'faction_pvp_future_war_cult': 'Future War Cult',
            'faction_pvp_new_monarchy': 'New Monarchy'
        };

        var currencyMap = {
            2033897755: 1415355184, // crucible marks
            2033897742: 1415355173 // vanguard marks
        };

        var statAttack = '368428387';
        var statDefense = '3897883278';
        var statLight = '2391494160';
        var factionFilterCache = DSCacheFactory('app.character.service:faction-filter', { storageMode: 'localStorage' });

        /**
         * True value means show the faction. False value means hide it.
         * @param id
         * @param value
         */
        this.setFactionFilter = function(id, value) {
            factionFilterCache.put(id, value);
        };

        /**
         * True value means show the faction. False value means hide it.
         * @param id
         * @return bool
         */
        this.getFactionFilter = function(id) {
            var result = factionFilterCache.get(id);
            if (result === undefined) {
                return true;
            }
            return result;
        };

        this.getFactionTagMap = function() {
            return factionTagMap;
        };

        this.filterItems = function(items, definitions) {
            var filtered = [];

            angular.forEach(items, function(item) {
                var definition = definitions.items[item.itemHash];
                var bucket = definitions.buckets[definition.bucketTypeHash];
                var talentGrid = definitions.talentGrids[item.talentGridHash];

                var object = {
                    id: definition.itemHash,
                    name: definition.itemName,
                    icon: definition.icon,
                    value: item.value,
                    damageType: item.damageType,
                    tierType: definition.tierType,
                    typeName: definition.itemTypeName,
                    bucketName: bucket.bucketName,
                    bucketOrder: bucket.bucketOrder,
                    light: false,
                    attack: false,
                    upgradesRemaining: false,
                    upgradesCompleted: false
                };

                if (talentGrid && item.progression) {
                    var progressionLevel = item.progression.level;
                    var upgradesRemaining = 0;
                    var upgradesCompleted = 0;

                    // Pre-load all unlocked nodes so we can check exclusive node groups later with 100% accuracy
                    var unlocked = [];
                    for (var n in talentGrid.nodes) {
                        if (item.nodes[n].isActivated) {
                            unlocked.push(talentGrid.nodes[n].nodeHash);
                        }
                    }

                    for (var n in talentGrid.nodes) {
                        var node = talentGrid.nodes[n];

                        if (node.column < 0 || node.row < 0) {
                            continue;
                        }

                        var exActivated = false;
                        for (var ex in node.exlusiveWithNodes) {
                            if (unlocked.indexOf(node.exlusiveWithNodes[ex]) > -1) {
                                exActivated = true;
                                break;
                            }
                        }

                        var step = node.steps[item.nodes[n].stepIndex];
                        if (step.activationRequirement.gridLevel > progressionLevel) {
                            upgradesRemaining++;
                        } else if (step.activationRequirement.gridLevel <= progressionLevel && !exActivated && !item.nodes[n].isActivated) {
                            upgradesRemaining++;
                        } else {
                            upgradesCompleted++;
                        }
                    }

                    object.upgradesRemaining = upgradesRemaining;
                    object.upgradesCompleted = upgradesCompleted;
                }

                for (var i in item.stats) {
                    var stat = item.stats[i];
                    if (stat.statHash == statLight) {
                        object.light = stat.value;
                        break;
                    }
                }

                if (item.primaryStat && item.primaryStat.statHash == statAttack) {
                    object.attack = item.primaryStat.value;
                }

                filtered.push(object);
            }.bind(this));
            return filtered;
        };

        this.filterCurrencies = function(progressions, currencies, definitions) {
            var items = this.filterItems(currencies, definitions);

            angular.forEach(items, function(item) {
                for (var i in progressions) {
                    if (currencyMap[progressions[i].id] == item.id) {
                        item.progressCurrent = progressions[i].level;
                        item.progressMax = 100;
                    }
                }
            }.bind(this));

            return items;
        };

        this.filterEquippable = function(equippable, definitions) {
            var filtered = [];

            angular.forEach(equippable, function(items) {
                if (!items.items[0]) {
                    return;
                }

                filtered.push(this.filterItems(items.items, definitions)[0]);
            }.bind(this));

            return filtered;
        };

        this.filterProgressions = function(progressions, definitions) {
            var filtered = [];

            angular.forEach(progressions, function(progression) {
                var definition = definitions.progressions[progression.progressionHash];

                var visible = true;
                if (!definition.icon || !this.getFactionFilter(definition.name)) {
                    visible = false;
                }

                filtered.push({
                    id: progression.progressionHash,
                    name: this.getFactionFromTag(definition.name),
                    level: progression.level,
                    progressToNextLevel: progression.progressToNextLevel,
                    nextLevelAt: progression.nextLevelAt,
                    icon: definition.icon,
                    numSteps: definition.steps.length,
                    repeatLastStep: definition.repeatLastStep,
                    visible: visible,
                    percent: progression.progressToNextLevel / progression.nextLevelAt,
                    maxed: progression.level == definition.steps.length && !definition.repeatLastStep
                });
            }.bind(this));
            return filtered;
        };

        this.filterWeeklyActivities = function(activities, definitions) {
            var filtered = [
                {
                    type: 'heroic',
                    name: 'Weekly Heroic Strike',
                    level: 22,
                    icon: 'weekly-strike.png',
                    complete: false
                },
                {
                    type: 'heroic',
                    name: 'Weekly Heroic Strike',
                    level: 26,
                    icon: 'weekly-strike.png',
                    complete: false
                },
                {
                    type: 'heroic',
                    name: 'Weekly Heroic Strike',
                    level: 28,
                    icon: 'weekly-strike.png',
                    complete: false
                },
                {
                    type: 'heroic',
                    name: 'Weekly Heroic Strike',
                    level: 30,
                    icon: 'weekly-strike.png',
                    complete: false
                },
                {
                    type: 'nightfall',
                    name: 'Weekly Nightfall Strike',
                    level: 30,
                    icon: 'weekly-strike.png',
                    complete: false
                },
                {
                    type: 'raid',
                    name: 'Vault of Glass',
                    level: 26,
                    icon: 'weekly-raid.png',
                    complete: false
                },
                {
                    type: 'raid',
                    name: 'Vault of Glass',
                    level: 30,
                    icon: 'weekly-raid.png',
                    complete: false
                },
                {
                    type: 'raid',
                    name: 'Crota\'s End',
                    level: 30,
                    icon: 'weekly-raid.png',
                    complete: false
                },
                {
                    type: 'raid',
                    name: 'Crota\'s End',
                    level: 33,
                    icon: 'weekly-raid.png',
                    complete: false
                }
            ];

            var tues = new Date();
            tues.setUTCHours(9,0,0,0);

            if (tues.getUTCDay() == 2 && new Date().getUTCHours() < 9) {
                tues.setDate(tues.getUTCDate() - 7);
            } else {
                var sinceTuesday = (tues.getUTCDay() - 2) < 0 ? (tues.getUTCDay() - 2) + 7 : tues.getUTCDay() - 2;
                tues.setDate(tues.getUTCDate() - sinceTuesday);
            }

            tues = tues.getTime();

            for (var i in activities) {
                var activity = activities[i];
                var date = new Date(activity.period).getTime();
                if (date < tues) {
                    break;
                }

                if (activity.values.completed.basic.value != 1 || activity.values.completionReason.basic.value != 0) {
                    continue;
                }

                var def = definitions.activities[activity.activityDetails.referenceId];
                for (var x in filtered) {
                    var name = def.activityName.trim().replace('’','\'');

                    if (filtered[x].name == name && filtered[x].level == def.activityLevel) {
                        filtered[x].complete = true;
                        break;
                    }
                }
            }

            return filtered;
        };

        this.getFactionFromTag = function(tag) {
            return factionTagMap[tag];
        };

        /**
         * @param platform
         * @param memberId
         * @param characterId
         * @returns {Animation.promise|promise|cd.g.promise|qFactory.Deferred.promise}
         */
        this.findCharacter = function(platform, memberId, characterId) {
            var d = $q.defer();
            var character;
            var definitions;

            // Don't worry, this result is cached by the BungieClient so the API isn't hit again.
            BungieClient.findCharacters(platform, memberId)
                .then(function(result) {
                    definitions = result.getDefinitions();

                    var characters = result.getData('characters');
                    for (var i = 0; i < characters.length; i++) {
                        if (characters[i].characterBase.characterId == characterId) {
                            character = characters[i];
                        }
                    }

                    d.resolve({
                        memberDetails: result.getData('memberDetails'),
                        definitions: definitions,
                        character: character
                    });
                }.bind(this), function(error) {
                    d.reject(error);
                });

            return d.promise;
        };
    });
